@@ -125,16 +125,54 @@
       this.scope = scope;
     }
 
+    function expression(stream, state) {
+      let match;
+      if (stream.match(/[[]/)) {
+        state.soyState.push("list-literal");
+        state.lookupVariables = false;
+        return null;
+      } else if (stream.match(/map\b/)) {
+        state.soyState.push("map-literal");
+        return "keyword";
+      } else if (stream.match(/record\b/)) {
+        state.soyState.push("record-literal");
+        return "keyword";
+      } else if (stream.match(/([\w]+)(?=\()/)) {
+        return "variable callee";
+      } else if (match = stream.match(/^["']/)) {
+        state.soyState.push("string");
+        state.quoteKind = match[0];
+        return "string";
+      } else if (stream.match(/^[(]/)) {
+        state.soyState.push("open-parentheses");
+        return null;
+      } else if (stream.match(/(null|true|false)(?!\w)/) ||
+          stream.match(/0x([0-9a-fA-F]{2,})/) ||
+          stream.match(/-?([0-9]*[.])?[0-9]+(e[0-9]*)?/)) {
+        return "atom";
+      } else if (stream.match(/(\||[+\-*\/%]|[=!]=|\?:|[<>]=?)/)) {
+        // Tokenize filter, binary, null propagator, and equality operators.
+        return "operator";
+      } else if (match = stream.match(/^\$([\w]+)/)) {
+        return ref(state.variables, match[1], !state.lookupVariables);
+      } else if (match = stream.match(/^\w+/)) {
+        return /^(?:as|and|or|not|in|if)$/.test(match[0]) ? "keyword" : null;
+      }
+
+      stream.next();
+      return null;
+    }
+
     return {
       startState: function() {
         return {
           soyState: [],
-          templates: null,
           variables: prepend(null, 'ij'),
           scopes: null,
           indent: 0,
           quoteKind: null,
           context: null,
+          lookupVariables: true, // Is unknown variables considered an error
           localStates: [{
             mode: modes.html,
             state: CodeMirror.startState(modes.html)
@@ -146,11 +184,11 @@
         return {
           tag: state.tag, // Last seen Soy tag.
           soyState: state.soyState.concat([]),
-          templates: state.templates,
           variables: state.variables,
           context: state.context,
           indent: state.indent, // Indentation of the following line.
           quoteKind: state.quoteKind,
+          lookupVariables: state.lookupVariables,
           localStates: state.localStates.map(function(localState) {
             return {
               mode: localState.mode,
@@ -202,7 +240,6 @@
         switch (last(state.soyState)) {
           case "templ-def":
             if (match = stream.match(/^\.?([\w]+(?!\.[\w]+)*)/)) {
-              state.templates = prepend(state.templates, match[1]);
               state.soyState.pop();
               return "def";
             }
@@ -248,18 +285,56 @@
             stream.next();
             return null;
 
-          case "param-type":
-            var peekChar = stream.peek();
-            if (peekChar == "}" || peekChar == "=") {
+          case "open-parentheses":
+            if (stream.match(/[)]/)) {
               state.soyState.pop();
               return null;
             }
-            if (stream.eatWhile(/^([\w]+|[?])/)) {
+            return expression(stream, state);
+
+          case "param-type":
+            var peekChar = stream.peek();
+            if ("}]=>,".indexOf(peekChar) != -1) {
+              state.soyState.pop();
+              return null;
+            } else if (peekChar == "[") {
+              state.soyState.push('param-type-record');
+              return null;
+            } else if (match = stream.match(/^([\w]+|[?])/)) {
+              if (match[0] == "map" || match[0] == "list") {
+                state.soyState.push('param-type-map-list');
+              }
               return "type";
             }
             stream.next();
             return null;
 
+          case "param-type-record":
+            var peekChar = stream.peek();
+            if (peekChar == "]") {
+              state.soyState.pop();
+              return null;
+            }
+            if (stream.match(/^\w+/)) {
+              state.soyState.push('param-type');
+              return "property";
+            }
+            stream.next();
+            return null;
+
+          case "param-type-map-list":
+            var peekChar = stream.peek();
+            if (stream.match(/^[>]/)) {
+              state.soyState.pop();
+              return null;
+            }
+            if (stream.match(/^[<,]/)) {
+              state.soyState.push('param-type');
+              return null;
+            }
+            stream.next();
+            return null;
+
           case "var-def":
             if (match = stream.match(/^\$([\w]+)/)) {
               state.variables = prepend(state.variables, match[1]);
@@ -269,6 +344,65 @@
             stream.next();
             return null;
 
+          case "record-literal":
+            if (stream.match(/^[)]/)) {
+              state.soyState.pop();
+              return null;
+            }
+            if (stream.match(/[(,]/)) {
+              state.soyState.push("map-value")
+              state.soyState.push("record-key")
+              return null;
+            }
+            stream.next()
+            return null;
+
+          case "map-literal":
+            if (stream.match(/^[)]/)) {
+              state.soyState.pop();
+              return null;
+            }
+            if (stream.match(/[(,]/)) {
+              state.soyState.push("map-value")
+              state.soyState.push("map-value")
+              return null;
+            }
+            stream.next()
+            return null;
+
+          case "list-literal":
+            if (stream.match(/\]/)) {
+              state.soyState.pop();
+              state.lookupVariables = true;
+              return null;
+            }
+            if (stream.match(/for\b/)) {
+              state.soyState.push("var-def")
+              return "keyword";
+            } else if (stream.match(/in\b/)) {
+              state.lookupVariables = true;
+              return "keyword";
+            }
+            return expression(stream, state);
+
+          case "record-key":
+            if (stream.match(/[\w]+/)) {
+              return "property";
+            }
+            if (stream.match(/^[:]/)) {
+              state.soyState.pop();
+              return null;
+            }
+            stream.next();
+            return null;
+            
+          case "map-value":
+            if (stream.peek() == ")" || stream.peek() == "," || stream.match(/^[:)]/)) {
+              state.soyState.pop();
+              return null;
+            }
+            return expression(stream, state);
+
           case "tag":
             var endTag = state.tag[0] == "/";
             var tagName = endTag ? state.tag.substring(1) : state.tag;
@@ -302,30 +436,8 @@
                 });
               }
               return "attribute";
-            } else if (match = stream.match(/([\w]+)(?=\()/)) {
-              return "variable callee";
-            } else if (match = stream.match(/^["']/)) {
-              state.soyState.push("string");
-              state.quoteKind = match;
-              return "string";
             }
-            if (stream.match(/(null|true|false)(?!\w)/) ||
-              stream.match(/0x([0-9a-fA-F]{2,})/) ||
-              stream.match(/-?([0-9]*[.])?[0-9]+(e[0-9]*)?/)) {
-              return "atom";
-            }
-            if (stream.match(/(\||[+\-*\/%]|[=!]=|\?:|[<>]=?)/)) {
-              // Tokenize filter, binary, null propagator, and equality operators.
-              return "operator";
-            }
-            if (match = stream.match(/^\$([\w]+)/)) {
-              return ref(state.variables, match[1]);
-            }
-            if (match = stream.match(/^\w+/)) {
-              return /^(?:as|and|or|not|in)$/.test(match[0]) ? "keyword" : null;
-            }
-            stream.next();
-            return null;
+            return expression(stream, state);
 
           case "literal":
             if (stream.match(/^(?=\{\/literal})/)) {
